#include "config.h"
#include "headers.h"

#define PRINTF (void)printf

#define UNSET	((unsigned int)0xab)

typedef union
{
    double f;
    unsigned char i[8];
} D;

#if defined(HAVE_LONG_DOUBLE)
typedef union
{
    long double f;
    unsigned char i[16];
} E;
#endif

void
show_nanf()
{
    float NaN;
    float zero;

    zero = (float)0.0;
    NaN = zero;
    NaN = zero/NaN;

    PRINTF("Test of NaN handling in single-precision IEEE 754 arithmetic:\n");
    PRINTF("\tsizeof(float) = %u\n", sizeof(float));

    PRINTF("\tNaN in hexadecimal = 0x%08x\n", (int)*((int*)&NaN));

    if (NaN == NaN) PRINTF("\t(NaN == NaN):\t\tWRONG\n");
    if (NaN != NaN) PRINTF("\t(NaN != NaN):\t\tOK\n");
    if (NaN <= NaN) PRINTF("\t(NaN <= NaN):\t\tWRONG\n");
    if (NaN <  NaN) PRINTF("\t(NaN <  NaN):\t\tWRONG\n");
    if (NaN >  NaN) PRINTF("\t(NaN >  NaN):\t\tWRONG\n");
    if (NaN >= NaN) PRINTF("\t(NaN >= NaN):\t\tWRONG\n");

    if (NaN == zero) PRINTF("\t(NaN == 0.0):\t\tWRONG\n");
    if (NaN != zero) PRINTF("\t(NaN != 0.0):\t\tOK\n");
    if (NaN <= zero) PRINTF("\t(NaN <= 0.0):\t\tWRONG\n");
    if (NaN <  zero) PRINTF("\t(NaN <  0.0):\t\tWRONG\n");
    if (NaN >  zero) PRINTF("\t(NaN >  0.0):\t\tWRONG\n");
    if (NaN >= zero) PRINTF("\t(NaN >= 0.0):\t\tWRONG\n");

    if (zero == NaN) PRINTF("\t(0.0 == NaN):\t\tWRONG\n");
    if (zero != NaN) PRINTF("\t(0.0 != NaN):\t\tOK\n");
    if (zero <= NaN) PRINTF("\t(0.0 <= NaN):\t\tWRONG\n");
    if (zero <  NaN) PRINTF("\t(0.0 <  NaN):\t\tWRONG\n");
    if (zero >  NaN) PRINTF("\t(0.0 >  NaN):\t\tWRONG\n");
    if (zero >= NaN) PRINTF("\t(0.0 >= NaN):\t\tWRONG\n");

    PRINTF("\n\n");
}


void
show_nan()
{
    double NaN;
    double zero;
    D x;
    int big_endian;
    int k;

    memset(&x.i, (int)UNSET, sizeof(x.i));
    x.f = (double)1.0;
    big_endian = (x.i[0] != 0);

    zero = (double)0.0;
    NaN = zero;
    NaN = zero/NaN;

    PRINTF("Test of NaN handling in double-precision IEEE 754 arithmetic:\n");
    PRINTF("\tsizeof(double) = %u\n", sizeof(double));

    x.f = NaN;
    PRINTF("\tNaN in hexadecimal = 0x");
    if (big_endian)
    {
	for (k = 0; k < sizeof(x.i); ++k)
	{
	    if (x.i[k] == UNSET)
		break;
	    if ((k % 4) == 0)
		PRINTF("_");
	    PRINTF("%02x",(unsigned int)x.i[k]);
	}
	PRINTF("\t[big-endian storage order]");
    }
    else
    {
	int n;

	for (k = sizeof(x.i) - 1; (k > 0) && (x.i[k] == UNSET); --k)
	    /* NO-OP */;

	for (n = 0; k >= 0; --k, ++n)
	{
	    if ((n % 4) == 0)
		PRINTF("_");
	    PRINTF("%02x",(unsigned int)x.i[k]);
	}
	PRINTF("\t[little-endian storage order]");
    }
    PRINTF("\n");

    if (NaN == NaN) PRINTF("\t(NaN == NaN):\t\tWRONG\n");
    if (NaN != NaN) PRINTF("\t(NaN != NaN):\t\tOK\n");
    if (NaN <= NaN) PRINTF("\t(NaN <= NaN):\t\tWRONG\n");
    if (NaN <  NaN) PRINTF("\t(NaN <  NaN):\t\tWRONG\n");
    if (NaN >  NaN) PRINTF("\t(NaN >  NaN):\t\tWRONG\n");
    if (NaN >= NaN) PRINTF("\t(NaN >= NaN):\t\tWRONG\n");

    if (NaN == zero) PRINTF("\t(NaN == 0.0):\t\tWRONG\n");
    if (NaN != zero) PRINTF("\t(NaN != 0.0):\t\tOK\n");
    if (NaN <= zero) PRINTF("\t(NaN <= 0.0):\t\tWRONG\n");
    if (NaN <  zero) PRINTF("\t(NaN <  0.0):\t\tWRONG\n");
    if (NaN >  zero) PRINTF("\t(NaN >  0.0):\t\tWRONG\n");
    if (NaN >= zero) PRINTF("\t(NaN >= 0.0):\t\tWRONG\n");

    if (zero == NaN) PRINTF("\t(0.0 == NaN):\t\tWRONG\n");
    if (zero != NaN) PRINTF("\t(0.0 != NaN):\t\tOK\n");
    if (zero <= NaN) PRINTF("\t(0.0 <= NaN):\t\tWRONG\n");
    if (zero <  NaN) PRINTF("\t(0.0 <  NaN):\t\tWRONG\n");
    if (zero >  NaN) PRINTF("\t(0.0 >  NaN):\t\tWRONG\n");
    if (zero >= NaN) PRINTF("\t(0.0 >= NaN):\t\tWRONG\n");

    PRINTF("\n\n");
}


void
show_nanl()
{
#if defined(HAVE_LONG_DOUBLE)
    long double NaN;
    long double zero;
    E x;
    int big_endian;
    int k;

    memset(&x.i, (int)UNSET, sizeof(x.i));
    x.f = (long double)1.0;
    big_endian = (x.i[0] != 0);

    zero = (long double)0.0;
    NaN = zero;
    NaN = zero/NaN;

    PRINTF("Test of NaN handling in extended-precision IEEE 754 arithmetic:\n");
    PRINTF("\tsizeof(long double) = %u\n", sizeof(long double));

    x.f = NaN;
    PRINTF("\tNaN in hexadecimal = 0x");
    if (big_endian)
    {
	for (k = 0; k < sizeof(x.i); ++k)
	{
	    if (x.i[k] == UNSET)
		break;
	    if ((k % 4) == 0)
		PRINTF("_");
	    PRINTF("%02x",(unsigned int)x.i[k]);
	}
	PRINTF("\t[big-endian storage order]");
    }
    else
    {
	int n;

	for (k = sizeof(x.i) - 1; (k > 0) && (x.i[k] == UNSET); --k)
	    /* NO-OP */;

	for (n = 0; k >= 0; --k, ++n)
	{
	    if ((n % 4) == 0)
		PRINTF("_");
	    PRINTF("%02x",(unsigned int)x.i[k]);
	}
	PRINTF("\t[little-endian storage order]");
    }
    PRINTF("\n");

    if (NaN == NaN) PRINTF("\t(NaN == NaN):\t\tWRONG\n");
    if (NaN != NaN) PRINTF("\t(NaN != NaN):\t\tOK\n");
    if (NaN <= NaN) PRINTF("\t(NaN <= NaN):\t\tWRONG\n");
    if (NaN <  NaN) PRINTF("\t(NaN <  NaN):\t\tWRONG\n");
    if (NaN >  NaN) PRINTF("\t(NaN >  NaN):\t\tWRONG\n");
    if (NaN >= NaN) PRINTF("\t(NaN >= NaN):\t\tWRONG\n");

    if (NaN == zero) PRINTF("\t(NaN == 0.0):\t\tWRONG\n");
    if (NaN != zero) PRINTF("\t(NaN != 0.0):\t\tOK\n");
    if (NaN <= zero) PRINTF("\t(NaN <= 0.0):\t\tWRONG\n");
    if (NaN <  zero) PRINTF("\t(NaN <  0.0):\t\tWRONG\n");
    if (NaN >  zero) PRINTF("\t(NaN >  0.0):\t\tWRONG\n");
    if (NaN >= zero) PRINTF("\t(NaN >= 0.0):\t\tWRONG\n");

    if (zero == NaN) PRINTF("\t(0.0 == NaN):\t\tWRONG\n");
    if (zero != NaN) PRINTF("\t(0.0 != NaN):\t\tOK\n");
    if (zero <= NaN) PRINTF("\t(0.0 <= NaN):\t\tWRONG\n");
    if (zero <  NaN) PRINTF("\t(0.0 <  NaN):\t\tWRONG\n");
    if (zero >  NaN) PRINTF("\t(0.0 >  NaN):\t\tWRONG\n");
    if (zero >= NaN) PRINTF("\t(0.0 >= NaN):\t\tWRONG\n");

    PRINTF("\n\n");
#endif
}


int
main(void)
{
    show_nanf();
    show_nan();
    show_nanl();

    return (EXIT_SUCCESS);
}
